from messageprocessorutil import UuidMessageProcessor
from messagebus import NakMessage, AckMessage


class OrderModifiedIgnoredTypesProcessor(UuidMessageProcessor):
    def __init__(self, ):
        # type: () -> None

        super(OrderModifiedIgnoredTypesProcessor, self).__init__()

    def parse_data(self, data, event_type):
        return None

    def call_business(self, _):
        return None

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_event(event, AckMessage())

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_event(event, NakMessage())

    def format_parse_exception(self, message_bus, message, event, data, exception):
        self.format_exception(message_bus, message, event, data, exception)
